﻿/*
 * Copyright 2011 Arne Vajhøj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

using System;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;

using Microsoft.CSharp;

namespace Vajhoej.StaticWeaver.Implementation
{
    /// <summary>
    /// Class MethodList contains mapping from signature to list of advices plus additional
    /// information (static or instance, assemblies needed for references).
    /// </summary>
    public class MethodList
    {
        private Dictionary<Signature, IList<AdviceInfo>> data = new Dictionary<Signature, IList<AdviceInfo>>();
        private Dictionary<Signature, bool> flag = new Dictionary<Signature, bool>();
        private List<Assembly> inpasm = new List<Assembly>();
        private List<Assembly> aspasm = new List<Assembly>();
        private Assembly glueasm;
        /// <summary>
        /// Add list of advices to signature.
        /// </summary>
        /// <param name="sig">Signature</param>
        /// <param name="adv">List of advices</param>
        public void Add(Signature sig, IList<AdviceInfo> adv)
        {
            data.Add(sig, adv);
        }
        /// <summary>
        /// Is signature already in list.
        /// </summary>
        /// <param name="sig">Signature</param>
        /// <returns>True if signature in list, false if signature not in list</returns>
        public bool Contains(Signature sig)
        {
            return data.ContainsKey(sig);
        }
        /// <summary>
        /// Get list of all signatures.
        /// </summary>
        /// <returns>List of signatures</returns>
        public IList<Signature> GetAll()
        {
            return data.Keys.ToList();
        }
        /// <summary>
        /// Get list of advices for signature.
        /// </summary>
        /// <param name="sig">Signature</param>
        /// <returns>List of advices</returns>
        public IList<AdviceInfo> GetAdvice(Signature sig)
        {
            return data[sig];
        }
        /// <summary>
        /// Mark signature as being static.
        /// </summary>
        /// <param name="sig">Signature</param>
        /// <param name="staticflag">True if static, false if instance</param>
        public void MarkStatic(Signature sig, bool staticflag)
        {
            flag[sig] = staticflag;
        }
        /// <summary>
        /// Is signature static.
        /// </summary>
        /// <param name="sig">Signature</param>
        /// <returns>True if static, false if instance</returns>
        public bool IsStatic(Signature sig)
        {
            return flag[sig];
        }
        /// <summary>
        /// Add input assembly to reference list.
        /// </summary>
        /// <param name="asmfnm">Input assembly filename</param>
        public void SetInputAssembly(string asmfnm)
        {
            inpasm.Add(Assembly.LoadFrom(asmfnm));
        }
        /// <summary>
        /// Add aspect assembly to reference list.
        /// </summary>
        /// <param name="asmfnm">Aspect assembly filename</param>
        public void SetAspectAssembly(string asmfnm)
        {
            aspasm.Add(Assembly.LoadFrom(asmfnm));
        }
        /// <summary>
        /// Add glue assembly to reference list.
        /// </summary>
        /// <param name="asmfnm">Glue assembly filename</param>
        public void SetGlueAssembly(string asmfnm)
        {
            glueasm = Assembly.LoadFrom(asmfnm);
        }
        /// <summary>
        /// Load type from one of input assemblies or glue assembly.
        /// </summary>
        /// <param name="typnam">Type name</param>
        /// <returns>Type</returns>
        public Type GetType(string typnam)
        {
            Type t = Type.GetType(typnam);
            if(t == null)
            {
                foreach(Assembly ia in inpasm)
                {
                    t = ia.GetType(typnam);
                    if(t != null) break;
                }
            }
            if(t == null && glueasm != null)
            {
                t = glueasm.GetType(typnam);
            }
            return t;
        }
        /// <summary>
        /// Get list of necesarry assemblies to reference when building glue assembly
        /// (input assemblies, aspect assemblies and StatisWeaver assembly).
        /// </summary>
        /// <returns></returns>
        public string[] GetNecesarryAssemblies()
        {
            List<string> res = new List<string>();
            foreach(Assembly ia in inpasm)
            {
                res.Add(ia.Location);
                foreach(string an in ia.GetReferencedAssemblies().Select(an => an.Name))
                {
                    if(File.Exists(an + ".exe"))
                    {
                        res.Add(an + ".exe");
                    }
                    else
                    {
                        res.Add(an + ".dll");
                    }
                }
            }
            foreach(Assembly aa in aspasm)
            {
                if(!res.Contains(aa.Location))
                {
                    res.Add(aa.Location);
                    foreach(string an in aa.GetReferencedAssemblies().Select(an => an.Name))
                    {
                        if(File.Exists(an + ".exe"))
                        {
                            res.Add(an + ".exe");
                        }
                        else
                        {
                            res.Add(an + ".dll");
                        }
                    }
                }
            }
            res.Add("StaticWeaver.dll");
            return res.ToArray();;
        }
    }
    /// <summary>
    /// Class CodeGen generates the glue assembly code.
    /// Note: this class uses .NET builtin compiler.
    /// </summary>
    public class CodeGen
    {
        private const string STATIC_VOID_METHOD_TEMPLATE = @"
namespace {0} {{
    public partial class {1}
    {{
        public static void {2}{14}({3})
        {{
            Vajhoej.StaticWeaver.MethodJoinpoint mjp = new Vajhoej.StaticWeaver.MethodJoinpoint(caller, null, ""{4}"", ""{5}""{9}{6});
            {7}
            {12}
            if(!mjp.SkipCall) {4}.{5}({11});
            {8}
            {13}
        }}
    }}
}}
";
        private const string STATIC_PROPSET_METHOD_TEMPLATE = @"
namespace {0} {{
    public partial class {1}
    {{
        public static void {2}{14}({3})
        {{
            Vajhoej.StaticWeaver.MethodJoinpoint mjp = new Vajhoej.StaticWeaver.MethodJoinpoint(caller, null, ""{4}"", ""{5}""{9}{6});
            {7}
            {12}
            if(!mjp.SkipCall) {4}.{5} = {11};
            {8}
            {13}
        }}
    }}
}}
";
        private const string STATIC_NONVOID_METHOD_TEMPLATE = @"
namespace {0} {{
    public partial class {1}
    {{
        public static {10} {2}{14}({3})
        {{
            Vajhoej.StaticWeaver.MethodJoinpoint mjp = new Vajhoej.StaticWeaver.MethodJoinpoint(caller, null, ""{4}"", ""{5}""{9}{6});
            {7}
            {12}
            if(!mjp.SkipCall) mjp.Result = {4}.{5}({11});
            {8}
            {13}
            return ({10})mjp.Result;
        }}
    }}
}}
";
        private const string STATIC_PROPGET_METHOD_TEMPLATE = @"
namespace {0} {{
    public partial class {1}
    {{
        public static {10} {2}{14}({3})
        {{
            Vajhoej.StaticWeaver.MethodJoinpoint mjp = new Vajhoej.StaticWeaver.MethodJoinpoint(caller, null, ""{4}"", ""{5}""{9}{6});
            {7}
            {12}
            if(!mjp.SkipCall) mjp.Result = {4}.{5};
            {8}
            {13}
            return ({10})mjp.Result;
        }}
    }}
}}
";
        private const string INSTANCE_VOID_METHOD_TEMPLATE = @"
namespace {0} {{
    public partial class {1}
    {{
        public static void {2}{14}({4} obj, {3})
        {{
            Vajhoej.StaticWeaver.MethodJoinpoint mjp = new Vajhoej.StaticWeaver.MethodJoinpoint(caller, obj, ""{4}"", ""{5}""{9}{6});
            {7}
            {12}
            if(!mjp.SkipCall) obj.{5}({11});
            {8}
            {13}
        }}
    }}
}}
";
        private const string INSTANCE_PROPSET_METHOD_TEMPLATE = @"
namespace {0} {{
    public partial class {1}
    {{
        public static void {2}{14}({4} obj, {3})
        {{
            Vajhoej.StaticWeaver.MethodJoinpoint mjp = new Vajhoej.StaticWeaver.MethodJoinpoint(caller, obj, ""{4}"", ""{5}""{9}{6});
            {7}
            {12}
            if(!mjp.SkipCall) obj.{5} = {11};
            {8}
            {13}
        }}
    }}
}}
";
        private const string INSTANCE_NONVOID_METHOD_TEMPLATE = @"
namespace {0} {{
    public partial class {1}
    {{
        public static {10} {2}{14}({4} obj, {3})
        {{
            Vajhoej.StaticWeaver.MethodJoinpoint mjp = new Vajhoej.StaticWeaver.MethodJoinpoint(caller, obj, ""{4}"", ""{5}""{9}{6});
            {7}
            {12}
            if(!mjp.SkipCall) mjp.Result = obj.{5}({11});
            {8}
            {13}
            return ({10})mjp.Result;
        }}
    }}
}}
";
        private const string INSTANCE_PROPGET_METHOD_TEMPLATE = @"
namespace {0} {{
    public partial class {1}
    {{
        public static {10} {2}{14}({4} obj, {3})
        {{
            Vajhoej.StaticWeaver.MethodJoinpoint mjp = new Vajhoej.StaticWeaver.MethodJoinpoint(caller, obj, ""{4}"", ""{5}""{9}{6});
            {7}
            {12}
            if(!mjp.SkipCall) mjp.Result = obj.{5};
            {8}
            {13}
            return ({10})mjp.Result;
        }}
    }}
}}
";
        private const string TRY_TEMPLATE = @"
            try
            {{
";
        private const string CATCH_TEMPLATE = @"
            }}
            catch(System.Exception ex)
            {{
                mjp.CaugthException = ex;
                {0}
            }}
";
        private static string GenBeforeCalls(IList<AdviceInfo> adv)
        {
            StringBuilder sb = new StringBuilder();
            foreach(AdviceInfo ai in adv)
            {
                if(ai.Type == AdviceType.BeforeCall)
                {
                    sb.Append(ai.Method.DeclaringType.FullName);
                    sb.Append(".");
                    sb.Append(ai.Method.Name);
                    sb.Append("(mjp);");
                }
            }
            return sb.ToString();
        }
        private static string GenAfterCalls(IList<AdviceInfo> adv)
        {
            StringBuilder sb = new StringBuilder();
            foreach(AdviceInfo ai in adv)
            {
                if(ai.Type == AdviceType.AfterCall)
                {
                    sb.Append(ai.Method.DeclaringType.FullName);
                    sb.Append(".");
                    sb.Append(ai.Method.Name);
                    sb.Append("(mjp);");
                }
            }
            return sb.ToString();
        }
        private static string GenAtExceptions(IList<AdviceInfo> adv)
        {
            StringBuilder sb = new StringBuilder();
            foreach(AdviceInfo ai in adv)
            {
                if(ai.Type == AdviceType.AtException)
                {
                    sb.Append(ai.Method.DeclaringType.FullName);
                    sb.Append(".");
                    sb.Append(ai.Method.Name);
                    sb.Append("(mjp);");
                }
            }
            return sb.ToString();
        }
        private static string GenArgList(Signature sig)
        {
            StringBuilder sb = new StringBuilder();
            for(int i = 0; i < sig.ParameterTypes.Count; i++)
            {
                if(i > 0) sb.Append(",");
                sb.Append("(");
                sb.Append(Signature.Internal2Source(sig.ParameterTypes[i]));
                sb.Append(")mjp.Arguments[");
                sb.Append(i.ToString());
                sb.Append("]");
            }
            return sb.ToString();
        }
        private static string GenGenericParams(Type typ, MethodInfo met)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("<");
            foreach(Type m in typ.GetGenericArguments())
            {
                if(sb.Length > 1) sb.Append(",");
                sb.Append(m.Name);
            }
            foreach(Type m in met.GetGenericArguments())
            {
                if(sb.Length > 1) sb.Append(",");
                sb.Append(m.Name);
            }
            sb.Append(">");
            return sb.ToString();
        }
        private static string GenGenericParams(Type typ)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("<");
            foreach(Type m in typ.GetGenericArguments())
            {
                if(sb.Length > 1) sb.Append(",");
                sb.Append(m.Name);
            }
            sb.Append(">");
            return sb.ToString();
        }
        /// <summary>
        /// Generate code for one method.
        /// </summary>
        /// <param name="ml">Mapping from signature to advices</param>
        /// <param name="sig">Signature</param>
        /// <returns>String with method code</returns>
        public static string GenMethod(MethodList ml, Signature sig)
        {
            Type dotnettyp = ml.GetType(Signature.Cecil2NoGeneric(sig.TypeName));
            MethodInfo dotnetmet;
            string genparm;
            string clzgenparm;
            if(sig.TypeName.Contains("<"))
            {
                dotnetmet = dotnettyp.GetMethods().Where(m => m.Name == sig.MethodName && m.GetParameters().Length == sig.ParameterTypes.Count).First();
                genparm = GenGenericParams(dotnettyp, dotnetmet);
                clzgenparm = GenGenericParams(dotnettyp);
            }
            else
            {
                dotnetmet = dotnettyp.GetMethod(sig.MethodName, sig.ParameterDotnetTypes);    
                genparm = "";
                clzgenparm = "";
            }
            IList<AdviceInfo> adv = ml.GetAdvice(sig);
            Signature newsig = sig.Convert(true);
            string template;
            if(dotnetmet.IsStatic)
            {
                ml.MarkStatic(sig, true);
                if(newsig.ReturnType == "System.Void")
                {
                    if(dotnetmet.IsSpecialName && sig.MethodName.StartsWith("set_"))
                    {
                        template = STATIC_PROPSET_METHOD_TEMPLATE;
                    }
                    else
                    {
                        template = STATIC_VOID_METHOD_TEMPLATE;
                    }
                }
                else
                {
                    if(dotnetmet.IsSpecialName && sig.MethodName.StartsWith("get_"))
                    {
                        template = STATIC_PROPGET_METHOD_TEMPLATE;
                    }
                    else
                    {
                        template = STATIC_NONVOID_METHOD_TEMPLATE;
                    }
                }
            }
            else
            {
                ml.MarkStatic(sig, false);
                if(newsig.ReturnType == "System.Void")
                {
                    if(dotnetmet.IsSpecialName && sig.MethodName.StartsWith("set_"))
                    {
                        template = INSTANCE_PROPSET_METHOD_TEMPLATE;
                    }
                    else
                    {
                        template = INSTANCE_VOID_METHOD_TEMPLATE;
                    }
                }
                else
                {
                    if(dotnetmet.IsSpecialName && sig.MethodName.StartsWith("get_"))
                    {
                        template = INSTANCE_PROPGET_METHOD_TEMPLATE;
                    }
                    else
                    {
                        template = INSTANCE_NONVOID_METHOD_TEMPLATE;
                    }
                }
            }
            string exadv = GenAtExceptions(adv);
            string trycode = exadv.Length > 0 ? String.Format(TRY_TEMPLATE) : "";
            string catchcode = exadv.Length > 0 ? String.Format(CATCH_TEMPLATE, exadv) : "";
            return String.Format(template,
                                 newsig.NameSpace, // 0
                                 newsig.ClassName, // 1
                                 newsig.MethodName, // 2
                                 Signature.Internal2Source(newsig.ParameterStringWithArgsInclCaller), // 3
                                 Signature.Internal2Source(clzgenparm.Length > 0 ? sig.TypeName.Split('<')[0] + clzgenparm : sig.TypeName), // 4
                                 dotnetmet.IsSpecialName ? sig.MethodName.Replace("get_", "").Replace("set_", "") : sig.MethodName, // 5
                                 sig.ArgsOnly, // 6
                                 GenBeforeCalls(adv), // 7
                                 GenAfterCalls(adv), // 8
                                 sig.ParameterTypes.Count > 0 ? ", " : "", // 9
                                 Signature.Internal2Source(newsig.ReturnType), // 10
                                 GenArgList(sig), // 11
                                 trycode, // 12
                                 catchcode, // 13
                                 genparm // 14
                                );
        }
        /// <summary>
        /// Generate code for all methods.
        /// </summary>
        /// <param name="ml">Mapping from signature to advices</param>
        /// <param name="glueassembly">Glue assembly filename</param></param>
        /// <param name="dbg">Debug flag (false =></param> no debug info, true => debug info written to Console.Out)</param>
        public static void Gen(MethodList ml, string glueassembly, bool dbg) 
        {
            StringBuilder sb = new StringBuilder();
            foreach(Signature sig in ml.GetAll())
            {
                sb.Append(GenMethod(ml, sig));
            }
            if(dbg)
            {
                Console.WriteLine(sb.ToString());
            }
            CodeDomProvider comp = new CSharpCodeProvider();
            CompilerParameters param = new CompilerParameters();
            param.OutputAssembly = glueassembly;
            param.ReferencedAssemblies.AddRange(ml.GetNecesarryAssemblies());
            CompilerResults res = comp.CompileAssemblyFromSource(param, sb.ToString());
            if(dbg && res.Errors.Count > 0)
            {
                foreach(CompilerError ce in res.Errors)
                {
                    Console.WriteLine(ce);
                }
            }
            ml.SetGlueAssembly(glueassembly);
        }
    }
}
